@@ -44,6 +44,7 @@
 #include "xua_as_fsm.h"
 #include "xua_asp_fsm.h"
 #include "xua_internal.h"
+#include "ss7_internal.h"
 
 #define M3UA_MSGB_SIZE 1500
 
@@ -448,7 +449,7 @@ int m3ua_decode_notify(struct osmo_xlm_prim_notify *npar, void *ctx,
 }
 
 /***********************************************************************
- * Transmitting M3UA messsages to SCTP
+ * Transmitting M3UA messages to SCTP
  ***********************************************************************/
 
 /* Convert M3UA from xua_msg to msgb and set PPID/stream */
@@ -511,7 +512,7 @@ int m3ua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 }
 
 /***********************************************************************
- * Receiving M3UA messsages from SCTP
+ * Receiving M3UA messages from SCTP
  ***********************************************************************/
 
 /* obtain the destination point code from a M3UA message in XUA fmt * */
@@ -553,6 +554,8 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	if (rc)
 		return rc;
 
+	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
+
 	/* FIXME: check for AS state == ACTIVE */
 
 	/* store the MTP-level information in the xua_msg for use by
@@ -691,7 +694,7 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 	OSMO_ASSERT(asp->cfg.proto == OSMO_SS7_ASP_PROT_M3UA);
 
 	/* caller owns msg memory, we shall neither free it here nor
-	 * keep references beyond the executin of this function and its
+	 * keep references beyond the execution of this function and its
 	 * callees */
 
 	xua = xua_from_msg(M3UA_VERSION, msgb_length(msg), msgb_data(msg));
@@ -752,8 +755,10 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 		err = m3ua_gen_error_msg(rc, msg);
 
 out:
-	if (err)
+	if (err) {
 		m3ua_tx_xua_asp(asp, err);
+		xua_msg_free(err);
+	}
 
 	xua_msg_free(xua);
 
@@ -818,9 +823,9 @@ static struct xua_msg *m3ua_encode_dupu(const uint32_t *rctx, unsigned int num_r
 					const char *info_string)
 {
 	struct xua_msg *xua = xua_msg_alloc();
-	uint32_t user_cause = (user << 16) | cause;
+	uint32_t user_cause = (cause << 16) | user;
 
-	xua->hdr = XUA_HDR(M3UA_MSGC_SNM, M3UA_SNM_DUNA);
+	xua->hdr = XUA_HDR(M3UA_MSGC_SNM, M3UA_SNM_DUPU);
 	xua->hdr.version = M3UA_VERSION;
 	if (rctx)
 		xua_msg_add_data(xua, M3UA_IEI_ROUTE_CTX, num_rctx * sizeof(*rctx), (const uint8_t *)rctx);
@@ -856,6 +861,7 @@ void m3ua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsig
 		xua = m3ua_encode_duna(rctx, num_rctx, aff_pc, num_aff_pc, info_string);
 
 	m3ua_tx_xua_asp(asp, xua);
+	xua_msg_free(xua);
 }
 
 /*! Transmit SSNM DUPU message indicating user unavailability.
@@ -871,6 +877,7 @@ void m3ua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int n
 {
 	struct xua_msg *xua = m3ua_encode_dupu(rctx, num_rctx, dpc, user, cause, info_str);
 	m3ua_tx_xua_asp(asp, xua);
+	xua_msg_free(xua);
 }
 
 /* received SNM message on ASP side */
@@ -901,6 +908,18 @@ static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 			xua->hdr.msg_type);
 		/* silently ignore those to not confuse the sender */
 		break;
+	case M3UA_SNM_DAUD:
+		/* RFC states only permitted in ASP->SG direction, not reverse. But some
+		 * equipment still sends it to us as ASP ?!? */
+		if (asp->cfg.quirks & OSMO_SS7_ASP_QUIRK_DAUD_IN_ASP) {
+			LOGPASP(asp, DLM3UA, LOGL_NOTICE, "quirk daud_in_asp active: Accepting DAUD "
+				"despite being in ASP role\n");
+			xua_snm_rx_daud(asp, xua);
+		} else {
+			LOGPASP(asp, DLM3UA, LOGL_ERROR, "DAUD not permitted in ASP role\n");
+			return M3UA_ERR_UNSUPP_MSG_TYPE;
+		}
+		break;
 	default:
 		return M3UA_ERR_UNSUPP_MSG_TYPE;
 	}
@@ -926,9 +945,15 @@ static int m3ua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	/* SNM only permitted in ACTIVE state */
 	if (asp->fi->state != XUA_ASP_S_ACTIVE) {
-		LOGPASP(asp, DLM3UA, LOGL_NOTICE, "Received M3UA SNM while ASP in state %s\n",
-			osmo_fsm_inst_state_name(asp->fi));
-		return M3UA_ERR_UNEXPECTED_MSG;
+		if (asp->fi->state == XUA_ASP_S_INACTIVE &&
+		    asp->cfg.quirks & OSMO_SS7_ASP_QUIRK_SNM_INACTIVE) {
+			LOGPASP(asp, DLM3UA, LOGL_NOTICE, "quirk snm_inactive active: "
+				"Accepting SNM in state %s\n", osmo_fsm_inst_state_name(asp->fi));
+		} else {
+			LOGPASP(asp, DLM3UA, LOGL_ERROR, "Rx M3UA SNM not permitted "
+				"while ASP in state %s\n", osmo_fsm_inst_state_name(asp->fi));
+			return M3UA_ERR_UNEXPECTED_MSG;
+		}
 	}
 
 	switch (asp->cfg.role) {
