@@ -48,6 +48,7 @@
 
 #include "sccp_internal.h"
 #include "xua_internal.h"
+#include "ss7_internal.h"
 #include "xua_asp_fsm.h"
 #include "xua_as_fsm.h"
 
@@ -196,6 +197,15 @@ uint8_t osmo_ss7_pc_width(const struct osmo_ss7_pc_fmt *pc_fmt)
 	return pc_fmt->component_len[0] + pc_fmt->component_len[1] + pc_fmt->component_len[2];
 }
 
+/* truncate pc or mask to maximum permitted length. This solves
+ * callers specifying arbitrary large masks which then evade duplicate
+ * detection with longer mask lengths */
+uint32_t osmo_ss7_pc_normalize(const struct osmo_ss7_pc_fmt *pc_fmt, uint32_t pc)
+{
+	uint32_t mask = (1 << osmo_ss7_pc_width(pc_fmt))-1;
+	return pc & mask;
+}
+
 /* get the number of bits we must shift the given component of a point
  * code in this ss7_instance */
 static unsigned int get_pc_comp_shift(const struct osmo_ss7_pc_fmt *pc_fmt,
@@ -605,7 +615,7 @@ osmo_ss7_linkset_find_or_create(struct osmo_ss7_instance *inst, const char *name
  * SS7 Link
  ***********************************************************************/
 
-/*! \brief Destryo SS7 Link
+/*! \brief Destroy SS7 Link
  *  \param[in] link SS7 Link to be destroyed */
 void osmo_ss7_link_destroy(struct osmo_ss7_link *link)
 {
@@ -702,6 +712,9 @@ osmo_ss7_route_find_dpc(struct osmo_ss7_route_table *rtbl, uint32_t dpc)
 	struct osmo_ss7_route *rt;
 
 	OSMO_ASSERT(ss7_initialized);
+
+	dpc = osmo_ss7_pc_normalize(&rtbl->inst->cfg.pc_fmt, dpc);
+
 	/* we assume the routes are sorted by mask length, i.e. more
 	 * specific routes first, and less specific routes with shorter
 	 * mask later */
@@ -720,6 +733,9 @@ osmo_ss7_route_find_dpc_mask(struct osmo_ss7_route_table *rtbl, uint32_t dpc,
 	struct osmo_ss7_route *rt;
 
 	OSMO_ASSERT(ss7_initialized);
+	mask = osmo_ss7_pc_normalize(&rtbl->inst->cfg.pc_fmt, mask);
+	dpc = osmo_ss7_pc_normalize(&rtbl->inst->cfg.pc_fmt, dpc);
+
 	/* we assume the routes are sorted by mask length, i.e. more
 	 * specific routes first, and less specific routes with shorter
 	 * mask later */
@@ -775,6 +791,11 @@ osmo_ss7_route_create(struct osmo_ss7_route_table *rtbl, uint32_t pc,
 	struct osmo_ss7_linkset *lset;
 	struct osmo_ss7_as *as = NULL;
 
+	/* truncate mask to maximum.  Let's avoid callers specifying arbitrary large
+	 * masks to ensure we don't fail duplicate detection with longer mask lengths */
+	mask = osmo_ss7_pc_normalize(&rtbl->inst->cfg.pc_fmt, mask);
+	pc = osmo_ss7_pc_normalize(&rtbl->inst->cfg.pc_fmt, pc);
+
 	OSMO_ASSERT(ss7_initialized);
 	lset = osmo_ss7_linkset_find_by_name(rtbl->inst, linkset_name);
 	if (!lset) {
@@ -783,6 +804,15 @@ osmo_ss7_route_create(struct osmo_ss7_route_table *rtbl, uint32_t pc,
 			return NULL;
 	}
 
+	/* check for duplicates */
+	rt = osmo_ss7_route_find_dpc_mask(rtbl, pc, mask);
+	if (rt && !strcmp(rt->cfg.linkset_name, linkset_name)) {
+		LOGSS7(rtbl->inst, LOGL_ERROR, "Refusing to create duplicate route: "
+			"pc=%u=%s mask=0x%x via linkset/AS '%s'\n",
+			pc, osmo_ss7_pointcode_print(rtbl->inst, pc), mask, linkset_name);
+		return rt;
+	}
+
 	rt = talloc_zero(rtbl, struct osmo_ss7_route);
 	if (!rt)
 		return NULL;
@@ -809,7 +839,13 @@ osmo_ss7_route_create(struct osmo_ss7_route_table *rtbl, uint32_t pc,
 /*! \brief Destroy a given SS7 route */
 void osmo_ss7_route_destroy(struct osmo_ss7_route *rt)
 {
+	struct osmo_ss7_route_table *rtbl = rt->rtable;
+
 	OSMO_ASSERT(ss7_initialized);
+
+	LOGSS7(rtbl->inst, LOGL_INFO, "Destroying route: pc=%u=%s mask=0x%x via linkset/ASP '%s'\n",
+	       rt->cfg.pc, osmo_ss7_pointcode_print(rtbl->inst, rt->cfg.pc), rt->cfg.mask, rt->cfg.linkset_name);
+
 	llist_del(&rt->list);
 	talloc_free(rt);
 }
@@ -860,6 +896,19 @@ const char *osmo_ss7_route_print(const struct osmo_ss7_route *rt)
  * SS7 Application Server
  ***********************************************************************/
 
+static const struct rate_ctr_desc ss7_as_rcd[] = {
+	[SS7_AS_CTR_RX_MSU_TOTAL] = { "rx:msu:total", "Total number of MSU received" },
+	[SS7_AS_CTR_TX_MSU_TOTAL] = { "tx:msu:total", "Total number of MSU transmitted" },
+};
+
+static const struct rate_ctr_group_desc ss7_as_rcgd = {
+	.group_name_prefix = "sigtran_as",
+	.group_description = "SIGTRAN Application Server",
+	.num_ctr = ARRAY_SIZE(ss7_as_rcd),
+	.ctr_desc = ss7_as_rcd,
+};
+static unsigned int g_ss7_as_rcg_idx;
+
 /*! \brief Find Application Server by given name
  *  \param[in] inst SS7 Instance on which we operate
  *  \param[in] name Name of AS
@@ -950,7 +999,7 @@ struct osmo_ss7_as *osmo_ss7_as_find_by_proto(struct osmo_ss7_instance *inst,
  *  \param[in] inst SS7 Instance on which we operate
  *  \param[in] name Name of Application Server
  *  \param[in] proto Protocol of Application Server
- *  \returns pointer to Application Server on suuccess; NULL otherwise */
+ *  \returns pointer to Application Server on success; NULL otherwise */
 struct osmo_ss7_as *
 osmo_ss7_as_find_or_create(struct osmo_ss7_instance *inst, const char *name,
 			   enum osmo_ss7_asp_protocol proto)
@@ -967,6 +1016,12 @@ osmo_ss7_as_find_or_create(struct osmo_ss7_instance *inst, const char *name,
 		as = talloc_zero(inst, struct osmo_ss7_as);
 		if (!as)
 			return NULL;
+		as->ctrg = rate_ctr_group_alloc(as, &ss7_as_rcgd, g_ss7_as_rcg_idx++);
+		if (!as->ctrg) {
+			talloc_free(as);
+			return NULL;
+		}
+		rate_ctr_group_set_name(as->ctrg, name);
 		as->inst = inst;
 		as->cfg.name = talloc_strdup(as, name);
 		as->cfg.proto = proto;
@@ -1056,6 +1111,7 @@ void osmo_ss7_as_destroy(struct osmo_ss7_as *as)
 
 	as->inst = NULL;
 	llist_del(&as->list);
+	rate_ctr_group_free(as->ctrg);
 	talloc_free(as);
 }
 
@@ -1090,6 +1146,26 @@ bool osmo_ss7_as_active(const struct osmo_ss7_as *as)
  * SS7 Application Server Process
  ***********************************************************************/
 
+enum ss7_asp_ctr {
+	SS7_ASP_CTR_PKT_RX_TOTAL,
+	SS7_ASP_CTR_PKT_RX_UNKNOWN,
+	SS7_ASP_CTR_PKT_TX_TOTAL,
+};
+
+static const struct rate_ctr_desc ss7_asp_rcd[] = {
+	[SS7_ASP_CTR_PKT_RX_TOTAL] = { "rx:packets:total", "Total number of packets received" },
+	[SS7_ASP_CTR_PKT_RX_UNKNOWN] = { "rx:packets:unknown", "Number of packets received for unknown PPID" },
+	[SS7_ASP_CTR_PKT_TX_TOTAL] = { "tx:packets:total", "Total number of packets transmitted" },
+};
+
+static const struct rate_ctr_group_desc ss7_asp_rcgd = {
+	.group_name_prefix = "sigtran_asp",
+	.group_description = "SIGTRAN Application Server Process",
+	.num_ctr = ARRAY_SIZE(ss7_asp_rcd),
+	.ctr_desc = ss7_asp_rcd,
+};
+static unsigned int g_ss7_asp_rcg_idx;
+
 int osmo_ss7_asp_peer_snprintf(char* buf, size_t buf_len, struct osmo_ss7_asp_peer *peer)
 {
 	int len = 0, offset = 0, rem = buf_len;
@@ -1126,7 +1202,7 @@ int osmo_ss7_asp_peer_snprintf(char* buf, size_t buf_len, struct osmo_ss7_asp_pe
  *  \param[in] talloc_ctx talloc context used to allocate new addresses.
  *  \param[in] hosts Array of strings containing IP addresses.
  *  \param[in] host_cnt Number of strings in hosts
- *  \returns 0 on success; negtive otherwise */
+ *  \returns 0 on success; negative otherwise */
 int osmo_ss7_asp_peer_set_hosts(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char* const* hosts, size_t host_cnt)
 {
 	int i = 0;
@@ -1156,12 +1232,11 @@ static inline bool host_is_ip_anyaddr(const char *host, bool is_v6)
 }
 
 /*! \brief Append (copy) address to a given ASP peer. Previous addresses are kept.
- *  \param[in] peer Application Server Process peer the address is appened to.
+ *  \param[in] peer Application Server Process peer the address is appended to.
  *  \param[in] talloc_ctx talloc context used to allocate new address.
  *  \param[in] host string containing an IP addresses.
- *  \returns 0 on success; negtive otherwise */
+ *  \returns 0 on success; negative otherwise */
 int osmo_ss7_asp_peer_add_host(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char *host)
-
 {
 	int i;
 	bool new_is_v6 = osmo_ip_str_type(host) == AF_INET6;
@@ -1191,6 +1266,73 @@ int osmo_ss7_asp_peer_add_host(struct osmo_ss7_asp_peer *peer, void *talloc_ctx,
 	return 0;
 }
 
+static bool ipv6_sctp_supported(const char *host, bool bind)
+{
+	int rc;
+	struct addrinfo hints;
+	struct addrinfo *result;
+	memset(&hints, 0, sizeof(struct addrinfo));
+	hints.ai_family = AF_INET6;
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_flags = AI_NUMERICHOST;
+	hints.ai_protocol = 0; /* Any protocol */
+
+	if (bind)  /* For wildcard IP address */
+		hints.ai_flags |= AI_PASSIVE;
+
+	/* man getaddrinfo: Either node or service, but not both, may be NULL. */
+	OSMO_ASSERT(host);
+	rc = getaddrinfo(host, NULL, &hints, &result);
+	if (rc != 0) {
+		LOGP(DLSS7, LOGL_NOTICE, "Default IPv6 address %s not supported: %s\n",
+		     host, gai_strerror(rc));
+		return false;
+	} else {
+		freeaddrinfo(result);
+		return true;
+	}
+}
+
+/* Set default values for local and remote peer hosts if they are not yet set.
+ *  \param[in] asp ASP for which to set default hosts.
+ *  \returns true if values where changed, false otherwise.
+ *
+ * If the ASP is already started, osmo_ss7_asp_restart() must be called
+ * afterwards in order to apply the new settings.
+ * This API is internal, hence doesn't appear in osmo_ss7.h
+ */
+bool osmo_ss7_asp_set_default_peer_hosts(struct osmo_ss7_asp *asp)
+{
+	bool changed = false;
+	/* If no local addr was set */
+	if (!asp->cfg.local.host_cnt) {
+		bool rem_has_v4 = false, rem_has_v6 = false;
+		int i;
+		for (i = 0; i < asp->cfg.remote.host_cnt; i++) {
+			if (osmo_ip_str_type(asp->cfg.remote.host[i]) == AF_INET6)
+				rem_has_v6 = true;
+			else
+				rem_has_v4 = true;
+		}
+		/* "::" Covers both IPv4 and IPv6, but if only IPv4
+		 * address are set on the remote side, IPv4 on the local
+		 * side must be set too */
+		if (ipv6_sctp_supported("::", true) && !(rem_has_v4 && !rem_has_v6))
+			osmo_ss7_asp_peer_add_host(&asp->cfg.local, asp, "::");
+		else
+			osmo_ss7_asp_peer_add_host(&asp->cfg.local, asp, "0.0.0.0");
+		changed = true;
+	}
+	/* If no remote addr was set */
+	if (!asp->cfg.remote.host_cnt) {
+		osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, "127.0.0.1");
+		if (ipv6_sctp_supported("::1", false))
+			osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, "::1");
+		changed = true;
+	}
+	return changed;
+}
+
 struct osmo_ss7_asp *
 osmo_ss7_asp_find_by_name(struct osmo_ss7_instance *inst, const char *name)
 {
@@ -1384,6 +1526,12 @@ osmo_ss7_asp_find_or_create(struct osmo_ss7_instance *inst, const char *name,
 	if (!asp) {
 		/* FIXME: check if local port has SCTP? */
 		asp = talloc_zero(inst, struct osmo_ss7_asp);
+		asp->ctrg = rate_ctr_group_alloc(asp, &ss7_asp_rcgd, g_ss7_asp_rcg_idx++);
+		if (!asp->ctrg) {
+			talloc_free(asp);
+			return NULL;
+		}
+		rate_ctr_group_set_name(asp->ctrg, name);
 		asp->inst = inst;
 		asp->cfg.remote.port = remote_port;
 		asp->cfg.local.port = local_port;
@@ -1427,6 +1575,7 @@ void osmo_ss7_asp_destroy(struct osmo_ss7_asp *asp)
 	/* unlink from ss7_instance */
 	asp->inst = NULL;
 	llist_del(&asp->list);
+	rate_ctr_group_free(asp->ctrg);
 	/* release memory */
 	talloc_free(asp);
 }
@@ -1680,6 +1829,8 @@ static int xua_srv_conn_cb(struct osmo_stream_srv *conn)
 	msgb_sctp_stream(msg) = sinfo.sinfo_stream;
 	msg->dst = asp;
 
+	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
+
 	if (ppid == SUA_PPID && asp->cfg.proto == OSMO_SS7_ASP_PROT_SUA)
 		rc = sua_rx_msg(asp, msg);
 	else if (ppid == M3UA_PPID && asp->cfg.proto == OSMO_SS7_ASP_PROT_M3UA)
@@ -1699,8 +1850,7 @@ static int xua_cli_connect_cb(struct osmo_stream_cli *cli)
 	struct osmo_ss7_asp *asp = osmo_stream_cli_get_data(cli);
 
 	/* update the socket name */
-	if (asp->sock_name)
-		talloc_free(asp->sock_name);
+	talloc_free(asp->sock_name);
 	asp->sock_name = osmo_sock_get_name(asp, ofd->fd);
 
 	LOGPASP(asp, DLSS7, LOGL_INFO, "Client connected %s\n", asp->sock_name);
@@ -1760,6 +1910,7 @@ static int ipa_cli_read_cb(struct osmo_stream_cli *conn)
 		return -1;
 	}
 	msg->dst = asp;
+	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
 	return ipa_rx_msg(asp, msg);
 }
 
@@ -1818,6 +1969,8 @@ static int xua_cli_read_cb(struct osmo_stream_cli *conn)
 	msgb_sctp_stream(msg) = sinfo.sinfo_stream;
 	msg->dst = asp;
 
+	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
+
 	if (ppid == SUA_PPID && asp->cfg.proto == OSMO_SS7_ASP_PROT_SUA)
 		rc = sua_rx_msg(asp, msg);
 	else if (ppid == M3UA_PPID && asp->cfg.proto == OSMO_SS7_ASP_PROT_M3UA)
@@ -1834,8 +1987,7 @@ static int xua_srv_conn_closed_cb(struct osmo_stream_srv *srv)
 {
 	struct osmo_ss7_asp *asp = osmo_stream_srv_get_data(srv);
 
-	LOGP(DLSS7, LOGL_INFO, "%s: SCTP connection closed\n",
-		asp ? asp->cfg.name : "?");
+	LOGP(DLSS7, LOGL_INFO, "%s: connection closed\n", asp ? asp->cfg.name : "?");
 
 	if (!asp)
 		return 0;
@@ -1953,8 +2105,7 @@ static int xua_accept_cb(struct osmo_stream_srv_link *link, int fd)
 	asp->server = srv;
 	asp->xua_server = oxs;
 	/* update the ASP socket name */
-	if (asp->sock_name)
-		talloc_free(asp->sock_name);
+	talloc_free(asp->sock_name);
 	asp->sock_name = talloc_reparent(link, asp, sock_name);
 	/* make sure the conn_cb() is called with the asp as private
 	 * data */
@@ -1988,6 +2139,8 @@ int osmo_ss7_asp_send(struct osmo_ss7_asp *asp, struct msgb *msg)
 		OSMO_ASSERT(0);
 	}
 
+	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_TX_TOTAL);
+
 	if (asp->cfg.is_server) {
 		if (!asp->server) {
 			LOGPASP(asp, DLSS7, LOGL_ERROR, "Cannot transmit, no asp->server\n");
@@ -2140,6 +2293,20 @@ osmo_ss7_xua_server_add_local_host(struct osmo_xua_server *xs, const char *local
 	return osmo_stream_srv_link_set_addrs(xs->server, (const char **)xs->cfg.local.host, xs->cfg.local.host_cnt);
 }
 
+bool osmo_ss7_xua_server_set_default_local_hosts(struct osmo_xua_server *oxs)
+{
+	/* If no local addr was set, or erased after _create(): */
+	if (!oxs->cfg.local.host_cnt) {
+		/* "::" Covers both IPv4 and IPv6 */
+		if (ipv6_sctp_supported("::", true))
+			osmo_ss7_xua_server_set_local_host(oxs, "::");
+		else
+			osmo_ss7_xua_server_set_local_host(oxs, "0.0.0.0");
+		return true;
+	}
+	return false;
+}
+
 void osmo_ss7_xua_server_destroy(struct osmo_xua_server *xs)
 {
 	struct osmo_ss7_asp *asp, *asp2;
@@ -2249,6 +2416,8 @@ static osmo_ss7_asp_rx_unknown_cb *g_osmo_ss7_asp_rx_unknown_cb;
 
 int ss7_asp_rx_unknown(struct osmo_ss7_asp *asp, int ppid_mux, struct msgb *msg)
 {
+	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_UNKNOWN);
+
 	if (g_osmo_ss7_asp_rx_unknown_cb)
 		return (*g_osmo_ss7_asp_rx_unknown_cb)(asp, ppid_mux, msg);
 
