@@ -1,6 +1,6 @@
 /* Core SS7 Instance/Linkset/Link/AS/ASP VTY Interface */
 
-/* (C) 2015-2017 by Harald Welte <laforge@gnumonks.org>
+/* (C) 2015-2021 by Harald Welte <laforge@gnumonks.org>
  * All Rights Reserved
  *
  * SPDX-License-Identifier: GPL-2.0+
@@ -37,14 +37,15 @@
 #include <osmocom/vty/telnet_interface.h>
 #include <osmocom/vty/misc.h>
 
+#include <osmocom/netif/stream.h>
+
 #include <osmocom/sigtran/osmo_ss7.h>
 #include <osmocom/sigtran/protocol/mtp.h>
 
 #include "xua_internal.h"
 #include <osmocom/sigtran/sccp_sap.h>
 #include "sccp_internal.h"
-
-#include "sccp_internal.h"
+#include "ss7_internal.h"
 
 #define XUA_VAR_STR	"(sua|m3ua|ipa)"
 
@@ -53,6 +54,19 @@
 	"MTP3 User Adaptation\n"	\
 	"IPA Multiplex (SCCP Lite)\n"
 
+static const struct value_string asp_quirk_names[] = {
+	{ OSMO_SS7_ASP_QUIRK_NO_NOTIFY,		"no_notify" },
+	{ OSMO_SS7_ASP_QUIRK_DAUD_IN_ASP,	"daud_in_asp" },
+	{ OSMO_SS7_ASP_QUIRK_SNM_INACTIVE,	"snm_inactive" },
+	{ 0, NULL }
+};
+
+static const struct value_string asp_quirk_descs[] = {
+	{ OSMO_SS7_ASP_QUIRK_NO_NOTIFY, "Peer SG doesn't send NTFY(AS-INACTIVE) after ASP-UP" },
+	{ OSMO_SS7_ASP_QUIRK_DAUD_IN_ASP, "Allow Rx of DAUD in ASP role" },
+	{ OSMO_SS7_ASP_QUIRK_SNM_INACTIVE, "Allow Rx of [S]SNM in AS-INACTIVE state" },
+	{ 0, NULL }
+};
 
 /***********************************************************************
  * Core CS7 Configuration
@@ -745,6 +759,36 @@ DEFUN_ATTR(asp_shutdown, asp_shutdown_cmd,
 	return CMD_WARNING;
 }
 
+DEFUN_ATTR(asp_quirk, asp_quirk_cmd,
+	"OVERWRITTEN",
+	"OVERWRITTEN\n",
+	CMD_ATTR_IMMEDIATE)
+{
+	struct osmo_ss7_asp *asp = vty->index;
+	int quirk = get_string_value(asp_quirk_names, argv[0]);
+
+	if (quirk < 0)
+		return CMD_WARNING;
+
+	asp->cfg.quirks |= quirk;
+	return CMD_SUCCESS;
+}
+
+DEFUN_ATTR(asp_no_quirk, asp_no_quirk_cmd,
+	"OVERWRITTEN",
+	"OVERWRITTEN\n",
+	CMD_ATTR_IMMEDIATE)
+{
+	struct osmo_ss7_asp *asp = vty->index;
+	int quirk = get_string_value(asp_quirk_names, argv[0]);
+
+	if (quirk < 0)
+		return CMD_WARNING;
+
+	asp->cfg.quirks &= ~quirk;
+	return CMD_SUCCESS;
+}
+
 DEFUN(show_cs7_asp, show_cs7_asp_cmd,
 	"show cs7 instance <0-15> asp",
 	SHOW_STR CS7_STR INST_STR INST_STR "Application Server Process (ASP)\n")
@@ -765,7 +809,15 @@ DEFUN(show_cs7_asp, show_cs7_asp_cmd,
 	vty_out(vty, "------------  ------------  -------------  ---- ----------------------- ----------%s", VTY_NEWLINE);
 
 	llist_for_each_entry(asp, &inst->asp_list, list) {
-		osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &asp->cfg.remote);
+		if (asp->cfg.proto == OSMO_SS7_ASP_PROT_IPA && asp->cfg.remote.port == 0 && asp->server) {
+			struct osmo_fd *ofd = osmo_stream_srv_get_ofd(asp->server);
+			char hostbuf[64];
+			char portbuf[16];
+			osmo_sock_get_ip_and_port(ofd->fd, hostbuf, sizeof(hostbuf),
+						  portbuf, sizeof(portbuf), false);
+			snprintf(buf, sizeof(buf), "%s:%s", hostbuf, portbuf);
+		} else
+			osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &asp->cfg.remote);
 		vty_out(vty, "%-12s  %-12s  %-13s  %-4s  %-14s  %-10s%s",
 			asp->cfg.name, "?",
 			asp->fi? osmo_fsm_inst_state_name(asp->fi) : "uninitialized",
@@ -804,6 +856,11 @@ static void write_one_asp(struct vty *vty, struct osmo_ss7_asp *asp, bool show_d
 	}
 	if (!asp->cfg.is_server)
 		vty_out(vty, "  sctp-role client%s", VTY_NEWLINE);
+	for (i = 0; i < sizeof(uint32_t) * 8; i++) {
+		if (!(asp->cfg.quirks & ((uint32_t) 1 << i)))
+			continue;
+		vty_out(vty, "  quirk %s%s", get_value_string(asp_quirk_names, (1 << i)), VTY_NEWLINE);
+	}
 }
 
 
@@ -992,6 +1049,7 @@ static int _rout_key(struct vty *vty,
 {
 	struct osmo_ss7_as *as = vty->index;
 	struct osmo_ss7_routing_key *rkey = &as->cfg.routing_key;
+	struct osmo_ss7_route *rt;
 	int pc;
 
 	if (as->cfg.proto == OSMO_SS7_ASP_PROT_IPA && atoi(rcontext) != 0) {
@@ -1005,12 +1063,33 @@ static int _rout_key(struct vty *vty,
 		vty_out(vty, "Invalid point code (%s)%s", dpc, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
+
+	/* When libosmo-sigtran is used in ASP role, the VTY routing table node
+	 * (config-cs7-rt) is not available. However, when we add a routing key
+	 * to an AS we still have to put a matching route into the routing
+	 * table. This is done automatically by first removing the old route
+	 * (users may change the routing key via VTY during runtime) and then
+	 * putting a new route (see below). */
+	if (cs7_role == CS7_ROLE_ASP) {
+		rt = osmo_ss7_route_find_dpc_mask(as->inst->rtable_system, rkey->pc, 0xffffff);
+		if (rt)
+			osmo_ss7_route_destroy(rt);
+	}
+
 	rkey->pc = pc;
 
 	rkey->context = atoi(rcontext);				/* FIXME: input validation */
 	rkey->si = si ? get_string_value(mtp_si_vals, si) : 0;	/* FIXME: input validation */
 	rkey->ssn = ssn ? atoi(ssn) : 0;			/* FIXME: input validation */
 
+	/* automatically add new route (see also comment above) */
+	if (cs7_role == CS7_ROLE_ASP) {
+		if (!osmo_ss7_route_create(as->inst->rtable_system, rkey->pc, 0xffffff, as->cfg.name)) {
+			vty_out(vty, "Cannot crate route (pc=%s, linkset=%s) to linkset %s", dpc, as->cfg.name, VTY_NEWLINE);
+			return CMD_WARNING;
+		}
+	}
+
 	return CMD_SUCCESS;
 }
 
@@ -1857,33 +1936,6 @@ static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst, bool
 		osmo_sccp_vty_write_cs7_node(vty, " ", inst->sccp);
 }
 
-static bool ipv6_sctp_supported(const char *host, bool bind)
-{
-	int rc;
-	struct addrinfo hints;
-	struct addrinfo *result;
-	memset(&hints, 0, sizeof(struct addrinfo));
-	hints.ai_family = AF_INET6;
-	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_flags = AI_NUMERICHOST;
-	hints.ai_protocol = 0; /* Any protocol */
-
-	if (bind)  /* For wildcard IP address */
-		hints.ai_flags |= AI_PASSIVE;
-
-	/* man getaddrinfo: Either node or service, but not both, may be NULL. */
-	OSMO_ASSERT(host);
-	rc = getaddrinfo(host, NULL, &hints, &result);
-	if (rc != 0) {
-		LOGP(DLSS7, LOGL_NOTICE, "Default IPv6 address %s not supported: %s\n",
-		     host, gai_strerror(rc));
-		return false;
-	} else {
-		freeaddrinfo(result);
-		return true;
-	}
-}
-
 int osmo_ss7_vty_go_parent(struct vty *vty)
 {
 	struct osmo_ss7_as *as;
@@ -1895,30 +1947,8 @@ int osmo_ss7_vty_go_parent(struct vty *vty)
 	switch (vty->node) {
 	case L_CS7_ASP_NODE:
 		asp = vty->index;
-		/* If no local addr was set */
-		if (!asp->cfg.local.host_cnt) {
-			bool rem_has_v4 = false, rem_has_v6 = false;
-			int i;
-			for (i = 0; i < asp->cfg.remote.host_cnt; i++) {
-				if (osmo_ip_str_type(asp->cfg.remote.host[i]) == AF_INET6)
-					rem_has_v6 = true;
-				else
-					rem_has_v4 = true;
-			}
-			/* "::" Covers both IPv4 and IPv6, but if only IPv4
-			 * address are set on the remote side, IPv4 on the local
-			 * side must be set too */
-			if (ipv6_sctp_supported("::", true) && !(rem_has_v4 && !rem_has_v6))
-				osmo_ss7_asp_peer_add_host(&asp->cfg.local, asp, "::");
-			else
-				osmo_ss7_asp_peer_add_host(&asp->cfg.local, asp, "0.0.0.0");
-		}
-		/* If no remote addr was set */
-		if (!asp->cfg.remote.host_cnt) {
-			osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, "127.0.0.1");
-			if (ipv6_sctp_supported("::1", false))
-				osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, "::1");
-		}
+		/* Make sure proper defaults values are set */
+		osmo_ss7_asp_set_default_peer_hosts(asp);
 		osmo_ss7_asp_restart(asp);
 		vty->node = L_CS7_NODE;
 		vty->index = asp->inst;
@@ -1936,13 +1966,7 @@ int osmo_ss7_vty_go_parent(struct vty *vty)
 	case L_CS7_XUA_NODE:
 		oxs = vty->index;
 		/* If no local addr was set, or erased after _create(): */
-		if (!oxs->cfg.local.host_cnt) {
-			/* "::" Covers both IPv4 and IPv6 */
-			if (ipv6_sctp_supported("::", true))
-				osmo_ss7_xua_server_set_local_host(oxs, "::");
-			else
-				osmo_ss7_xua_server_set_local_host(oxs, "0.0.0.0");
-		}
+		osmo_ss7_xua_server_set_default_local_hosts(oxs);
 		if (osmo_ss7_xua_server_bind(oxs) < 0)
 			vty_out(vty, "%% Unable to bind xUA server to IP(s)%s", VTY_NEWLINE);
 		vty->node = L_CS7_NODE;
@@ -2019,6 +2043,17 @@ static void vty_init_shared(void *ctx)
 {
 	g_ctx = ctx;
 
+	asp_quirk_cmd.string = vty_cmd_string_from_valstr(ctx, asp_quirk_names,
+							  "quirk (", "|", ")", VTY_DO_LOWER);
+	asp_quirk_cmd.doc = vty_cmd_string_from_valstr(ctx, asp_quirk_descs,
+							"Enable quirk to work around interop issues\n",
+							"\n", "\n", 0);
+	asp_no_quirk_cmd.string = vty_cmd_string_from_valstr(ctx, asp_quirk_names,
+							  "no quirk (", "|", ")", VTY_DO_LOWER);
+	asp_no_quirk_cmd.doc = vty_cmd_string_from_valstr(ctx, asp_quirk_descs,
+							NO_STR "Disable quirk to work around interop issues\n",
+							"\n", "\n", 0);
+
 	install_lib_element_ve(&show_cs7_user_cmd);
 	install_lib_element_ve(&show_cs7_xua_cmd);
 	install_lib_element_ve(&show_cs7_config_cmd);
@@ -2047,6 +2082,8 @@ static void vty_init_shared(void *ctx)
 	install_lib_element(L_CS7_ASP_NODE, &asp_sctp_role_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_block_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_shutdown_cmd);
+	install_lib_element(L_CS7_ASP_NODE, &asp_quirk_cmd);
+	install_lib_element(L_CS7_ASP_NODE, &asp_no_quirk_cmd);
 
 	install_node(&as_node, NULL);
 	install_lib_element_ve(&show_cs7_as_cmd);
@@ -2066,6 +2103,8 @@ static void vty_init_shared(void *ctx)
 	install_lib_element(L_CS7_AS_NODE, &as_pc_override_cmd);
 	install_lib_element(L_CS7_AS_NODE, &as_pc_patch_sccp_cmd);
 
+	install_lib_element_ve(&show_cs7_route_cmd);
+
 	vty_init_addr();
 }
 
@@ -2081,7 +2120,6 @@ void osmo_ss7_vty_init_sg(void *ctx)
 	vty_init_shared(ctx);
 
 	install_node(&rtable_node, NULL);
-	install_lib_element_ve(&show_cs7_route_cmd);
 	install_lib_element(L_CS7_NODE, &cs7_route_table_cmd);
 	install_lib_element(L_CS7_RTABLE_NODE, &cfg_description_cmd);
 	install_lib_element(L_CS7_RTABLE_NODE, &cs7_rt_upd_cmd);
