@@ -36,7 +36,7 @@
  * However, all SCCP features can be expressed in SUA.
  *
  * The code only supports Class 2.  No support for Class 3 is intended,
- * but patches are of course alwys welcome.
+ * but patches are of course always welcome.
  *
  * Missing other features:
  *  * Segmentation/Reassembly support
@@ -80,7 +80,7 @@ struct sccp_connection {
 	/* remote point code */
 	uint32_t remote_pc;
 
-	/* local/remote addresses and identiies */
+	/* local/remote addresses and identities */
 	struct osmo_sccp_addr calling_addr;
 	struct osmo_sccp_addr called_addr;
 	/* SCCP connection identifier. Only relevant across the SCCP User SAP,
@@ -90,7 +90,7 @@ struct sccp_connection {
 	/* SCCP Remote Connection Reference.  Allocated by the remote
 	 * SCCP stack to uniquely identify a SCCP connection on its end.
 	 * We don't interpret it, but simply cache it here so we can use
-	 * it whever sending data to the peer. Only relevant over the
+	 * it whenever sending data to the peer. Only relevant over the
 	 * wire, not to be used across the SCCP user SAP */
 	uint32_t remote_ref;
 
@@ -499,7 +499,10 @@ static struct sccp_connection *conn_create(struct osmo_sccp_user *user)
 	uint32_t conn_id;
 
 	do {
-		conn_id = user->inst->next_id++;
+		/* modulo 2^24 as we currently use the connection ID also as local
+		 * reference, and that is limited to 24 bits */
+		user->inst->next_id = (user->inst->next_id + 1) %  (1 << 24);
+		conn_id = user->inst->next_id;
 	} while (conn_find_by_id(user->inst, conn_id));
 
 	return conn_create_id(user, conn_id);
@@ -734,6 +737,7 @@ static void scu_gen_encode_and_send(struct sccp_connection *conn, uint32_t event
 		udisp = &scu_prim->u.disconnect;
 		udisp->conn_id = conn->conn_id;
 		udisp->responding_addr = conn->called_addr;
+		udisp->importance = conn->importance;
 		udisp->originator = OSMO_SCCP_ORIG_UNDEFINED;
 		//udisp->in_sequence_control;
 		if (xua) {
@@ -1080,11 +1084,19 @@ static void scoc_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *data
 		osmo_fsm_inst_state_chg(fi, S_IDLE, 0, 0);
 		break;
 	case SCOC_E_T_IAR_EXP:
+		/* stop inact timers */
+		conn_stop_inact_timers(conn);
+		xua = xua_msg_alloc();
+		xua_msg_add_u32(xua, SUA_IEI_CAUSE,
+				SUA_CAUSE_T_RELEASE | SCCP_RELEASE_CAUSE_EXPIRATION_INACTIVE);
+		xua_msg_add_u32(xua, SUA_IEI_IMPORTANCE, conn->importance);
 		/* Send N-DISCONNECT.ind to local user */
-		scu_gen_encode_and_send(conn, event, NULL, OSMO_SCU_PRIM_N_DISCONNECT,
+		scu_gen_encode_and_send(conn, event, xua, OSMO_SCU_PRIM_N_DISCONNECT,
 					PRIM_OP_INDICATION);
+		talloc_free(xua);
 		/* Send RLSD to peer */
 		xua_gen_relre_and_send(conn, SCCP_RELEASE_CAUSE_EXPIRATION_INACTIVE, NULL);
+		/* start release timer */
 		conn_start_rel_timer(conn);
 		osmo_fsm_inst_state_chg(fi, S_DISCONN_PEND, 0, 0);
 		break;
@@ -1120,14 +1132,20 @@ static void scoc_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *data
 		if (xua_msg_get_u32(xua, SUA_IEI_SRC_REF) != conn->remote_ref ||
 		    xua_msg_get_u32(xua, SUA_IEI_PROTO_CLASS) != conn->sccp_class) {
 			/* Release connection */
-			/* send N-DISCONNECT.ind to user */
-			scu_gen_encode_and_send(conn, event, NULL,
-						OSMO_SCU_PRIM_N_DISCONNECT,
-						PRIM_OP_INDICATION);
 			/* Stop inactivity Timers */
 			conn_stop_inact_timers(conn);
+			xua = xua_msg_alloc();
+			xua_msg_add_u32(xua, SUA_IEI_CAUSE,
+					SUA_CAUSE_T_RELEASE | SCCP_RELEASE_CAUSE_INCONSISTENT_CONN_DATA);
+			xua_msg_add_u32(xua, SUA_IEI_IMPORTANCE, conn->importance);
+			/* send N-DISCONNECT.ind to user */
+			scu_gen_encode_and_send(conn, event, xua,
+						OSMO_SCU_PRIM_N_DISCONNECT,
+						PRIM_OP_INDICATION);
+			talloc_free(xua);
 			/* Send RLSD to SCRC */
 			xua_gen_relre_and_send(conn, SCCP_RELEASE_CAUSE_INCONSISTENT_CONN_DATA, NULL);
+			talloc_free(xua);
 			/* Start release timer */
 			conn_start_rel_timer(conn);
 			osmo_fsm_inst_state_chg(fi, S_DISCONN_PEND, 0, 0);
@@ -1364,6 +1382,9 @@ void sccp_scoc_rx_scrc_rout_fail(struct osmo_sccp_instance *inst,
 	uint32_t conn_id;
 	struct sccp_connection *conn;
 
+	LOGP(DLSCCP, LOGL_NOTICE, "SCRC Routing Failure for message %s\n",
+	     xua_hdr_dump(xua, &xua_dialect_sua));
+
 	/* try to dispatch to connection FSM (if any) */
 	conn_id = xua_msg_get_u32(xua, SUA_IEI_DEST_REF);
 	conn = conn_find_by_id(inst, conn_id);
@@ -1500,7 +1521,7 @@ static void tx_rlsd_from_xua_twoway(struct sccp_connection *conn,
 	xua_msg_free(xua);
 }
 
-/* process received message for unasigned local reference */
+/* process received message for unassigned local reference */
 static void sccp_scoc_rx_unass_local_ref(struct osmo_sccp_instance *inst,
 					 struct xua_msg *xua)
 {
@@ -1610,7 +1631,7 @@ void sccp_scoc_rx_from_scrc(struct osmo_sccp_instance *inst,
 	/* we basically try to convert the SUA message into an event,
 	 * and then dispatch the event to the connection-specific FSM.
 	 * If it is a CORE (Connect REquest), we create the connection
-	 * (and imlpicitly its FSM) first */
+	 * (and implicitly its FSM) first */
 
 	if (xua->hdr.msg_type == SUA_CO_CORE) {
 		scu = sccp_find_user(inst, xua);
@@ -1631,8 +1652,8 @@ void sccp_scoc_rx_from_scrc(struct osmo_sccp_instance *inst,
 		conn_id = xua_msg_get_u32(xua, SUA_IEI_DEST_REF);
 		conn = conn_find_by_id(inst, conn_id);
 		if (!conn) {
-			LOGP(DLSCCP, LOGL_NOTICE, "Cannot find connection for "
-			     "local reference %u\n", conn_id);
+			LOGP(DLSCCP, LOGL_NOTICE, "Received %s: Cannot find connection for "
+			     "local reference %u\n", xua_hdr_dump(xua, &xua_dialect_sua), conn_id);
 			sccp_scoc_rx_unass_local_ref(inst, xua);
 			return;
 		}
