@@ -46,6 +46,7 @@
 #include "xua_asp_fsm.h"
 #include "xua_internal.h"
 #include "sccp_internal.h"
+#include "ss7_internal.h"
 
 /* Appendix C.4 of Q.714 (all in milliseconds) */
 #define CONNECTION_TIMER	( 1 * 60 * 100)
@@ -521,6 +522,8 @@ static int sua_rx_cl(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	if (rc)
 		return rc;
 
+	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
+
 	switch (xua->hdr.msg_type) {
 	case 0: /* Reserved, permitted by ETSI TS 101 592 5.2.3.2 */
 	case SUA_CL_CLDT:
@@ -549,6 +552,8 @@ static int sua_rx_co(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	if (rc)
 		return rc;
 
+	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
+
 	switch (xua->hdr.msg_type) {
 	case 0: /* Reserved, permitted by ETSI TS 101 592 5.2.3.2 */
 	case SUA_CO_CORE:
@@ -912,6 +917,18 @@ static int sua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 			xua->hdr.msg_type);
 		/* silently ignore those to not confuse the sender */
 		break;
+	case SUA_SNM_DAUD:
+		/* RFC states only permitted in ASP->SG direction, not reverse. But some
+		 * equipment still sends it to us as ASP ?!? */
+		if (asp->cfg.quirks & OSMO_SS7_ASP_QUIRK_DAUD_IN_ASP) {
+			LOGPASP(asp, DLSUA, LOGL_NOTICE, "quirk daud_in_asp active: Accepting DAUD "
+				"despite being in ASP role\n");
+			xua_snm_rx_daud(asp, xua);
+		} else {
+			LOGPASP(asp, DLSUA, LOGL_ERROR, "DAUD not permitted in ASP role\n");
+			return SUA_ERR_UNSUPP_MSG_TYPE;
+		}
+		break;
 	default:
 		return SUA_ERR_UNSUPP_MSG_TYPE;
 	}
@@ -937,9 +954,14 @@ static int sua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	/* SNM only permitted in ACTIVE state */
 	if (asp->fi->state != XUA_ASP_S_ACTIVE) {
-		LOGPASP(asp, DLSUA, LOGL_NOTICE, "Received M3UA SNM while ASP in state %s\n",
-			osmo_fsm_inst_state_name(asp->fi));
-		return SUA_ERR_UNEXPECTED_MSG;
+		if (asp->fi->state == XUA_ASP_S_INACTIVE && asp->cfg.quirks & OSMO_SS7_ASP_QUIRK_SNM_INACTIVE) {
+			LOGPASP(asp, DLSUA, LOGL_NOTICE, "quirk snm_inactive active: "
+				"Accepting SSNM in state %s\n", osmo_fsm_inst_state_name(asp->fi));
+		} else {
+			LOGPASP(asp, DLM3UA, LOGL_ERROR, "Rx SUA SSNM not permitted "
+				"while ASP in state %s\n", osmo_fsm_inst_state_name(asp->fi));
+			return SUA_ERR_UNEXPECTED_MSG;
+		}
 	}
 
 	switch (asp->cfg.role) {
