@@ -29,6 +29,7 @@
 
 #include "xua_asp_fsm.h"
 #include "xua_as_fsm.h"
+#include "xua_internal.h"
 
 #define S(x)	(1 << (x))
 
@@ -337,6 +338,21 @@ static const uint32_t evt_ack_map[_NUM_XUA_ASP_E] = {
 	[XUA_ASP_E_ASPSM_BEAT] = XUA_ASP_E_ASPSM_BEAT_ACK,
 };
 
+/* Helper function to dispatch an ASP->AS event to all AS of which this
+ * ASP is a memmber.  Ignores routing contexts for now. */
+static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event)
+{
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	struct osmo_ss7_instance *inst = asp->inst;
+	struct osmo_ss7_as *as;
+
+	llist_for_each_entry(as, &inst->as_list, list) {
+		if (!osmo_ss7_as_has_asp(as, asp))
+			continue;
+		osmo_fsm_inst_dispatch(as->fi, event, asp);
+	}
+}
 
 /* check if expected message was received + stop t_ack */
 static void check_stop_t_ack(struct osmo_fsm_inst *fi, uint32_t event)
@@ -380,6 +396,16 @@ static void check_stop_t_ack(struct osmo_fsm_inst *fi, uint32_t event)
 		}							\
 	} while(0)
 
+
+/***************
+** FSM states **
+***************/
+
+static void xua_asp_fsm_down_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_DOWN_IND);
+}
+
 static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
@@ -435,25 +461,9 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	}
 }
 
-/* Helper function to dispatch an ASP->AS event to all AS of which this
- * ASP is a memmber.  Ignores routing contexts for now. */
-static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event)
+static void xua_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	struct xua_asp_fsm_priv *xafp = fi->priv;
-	struct osmo_ss7_asp *asp = xafp->asp;
-	struct osmo_ss7_instance *inst = asp->inst;
-	struct osmo_ss7_as *as;
-
-	llist_for_each_entry(as, &inst->as_list, list) {
-		if (!osmo_ss7_as_has_asp(as, asp))
-			continue;
-		osmo_fsm_inst_dispatch(as->fi, event, asp);
-	}
-}
-
-static void xua_asp_fsm_down_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
-{
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_DOWN_IND);
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
 }
 
 static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -570,9 +580,9 @@ static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 	}
 }
 
-static void xua_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+static void xua_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
 }
 
 static void xua_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -651,11 +661,6 @@ static void xua_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *d
 	}
 }
 
-static void xua_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
-{
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
-}
-
 static void xua_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
@@ -849,6 +854,10 @@ static int get_fd_from_iafp(struct ipa_asp_fsm_priv *iafp)
 	return ofd->fd;
 }
 
+/***************
+** FSM states **
+***************/
+
 /* Server + Client: Initial State, wait for M-ASP-UP.req */
 static void ipa_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
@@ -866,7 +875,7 @@ static void ipa_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 			}
 		} else {
 			/* Client: We simply wait for an ID GET */
-			osmo_fsm_inst_state_chg(fi, IPA_ASP_S_WAIT_ID_ACK, 10, T_WAIT_ID_ACK);
+			osmo_fsm_inst_state_chg(fi, IPA_ASP_S_WAIT_ID_GET, 10, T_WAIT_ID_GET);
 		}
 		break;
 	}
@@ -938,23 +947,11 @@ out_err:
 /* Server: We're waiting for an ID ACK */
 static void ipa_asp_fsm_wait_id_ack2(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
-	struct ipa_asp_fsm_priv *iafp = fi->priv;
-	struct osmo_ss7_asp *asp = iafp->asp;
-	struct osmo_ss7_instance *inst = asp->inst;
-	/* We use routing-context '0' here, as that's the only one we support in IPA */
-	struct osmo_ss7_as *as = osmo_ss7_as_find_by_rctx(inst, 0);
-
-	OSMO_ASSERT(as);
-
 	switch (event) {
 	case IPA_ASP_E_ID_ACK:
 		/* ACK received, we can go to active state now.  The
 		 * ACTIVE onenter function will inform the AS */
 		osmo_fsm_inst_state_chg(fi, IPA_ASP_S_ACTIVE, 0, 0);
-		/* As opposed to M3UA, there is no RKM and we have to implicitly automatically add
-		 * a route once an IPA connection has come up */
-		osmo_ss7_route_create(inst->rtable_system, as->cfg.routing_key.pc, 0xffffff,
-				      as->cfg.name);
 		break;
 	}
 }
@@ -967,6 +964,7 @@ static void ipa_asp_fsm_wait_id_get(struct osmo_fsm_inst *fi, uint32_t event, vo
 	struct msgb *msg_get, *msg_resp;
 	const uint8_t *req_data;
 	int data_len;
+	int fd;
 
 	switch (event) {
 	case IPA_ASP_E_ID_GET:
@@ -989,6 +987,16 @@ static void ipa_asp_fsm_wait_id_get(struct osmo_fsm_inst *fi, uint32_t event, vo
 		osmo_ss7_asp_send(asp, msg_resp);
 		osmo_fsm_inst_state_chg(fi, IPA_ASP_S_WAIT_ID_ACK, 10, T_WAIT_ID_ACK);
 		break;
+	case IPA_ASP_E_ID_ACK:
+		/* Some SCCPLite MSCs are known to send an ACK directly instead
+		 * of GET. Support them and skip the GET+RESP handshake by
+		 * sending ACK2 to server directly */
+		fd = get_fd_from_iafp(iafp);
+		if (fd >= 0) {
+			ipaccess_send_id_ack(fd);
+			osmo_fsm_inst_state_chg(fi, IPA_ASP_S_ACTIVE, 0, 0);
+		}
+		break;
 	}
 }
 
@@ -1010,29 +1018,10 @@ static void ipa_asp_fsm_wait_id_ack(struct osmo_fsm_inst *fi, uint32_t event, vo
 	}
 }
 
-static void ipa_asp_fsm_del_route(struct ipa_asp_fsm_priv *iafp)
+static void ipa_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	struct osmo_ss7_asp *asp = iafp->asp;
-	struct osmo_ss7_instance *inst = asp->inst;
-	/* We use routing-context '0' here, as that's the only one we support in IPA */
-	struct osmo_ss7_as *as = osmo_ss7_as_find_by_rctx(inst, 0);
-	struct osmo_ss7_route *rt;
-
-	OSMO_ASSERT(as);
-
-	/* find the route which we have created if we ever reached ipa_asp_fsm_wait_id_ack2 */
-	rt = osmo_ss7_route_find_dpc_mask(inst->rtable_system, as->cfg.routing_key.pc, 0xffffff);
-	/* no route found, bail out */
-	if (!rt)
-		return;
-	/* route points to different AS, bail out */
-	if (rt->dest.as != as)
-		return;
-
-	osmo_ss7_route_destroy(rt);
-	/* FIXME: Why don't we also delete this timer if we return early above?
-	 * FIXME: Where is this timer even scheduled? */
-	osmo_timer_del(&iafp->pong_timer);
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
 }
 
 /* Server + Client: We're actively transmitting user data */
@@ -1041,17 +1030,20 @@ static void ipa_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *d
 	switch (event) {
 	case XUA_ASP_E_M_ASP_DOWN_REQ:
 	case XUA_ASP_E_M_ASP_INACTIVE_REQ:
-		ipa_asp_fsm_del_route(fi->priv);
 		osmo_fsm_inst_state_chg(fi, IPA_ASP_S_DOWN, 0, 0);
 		break;
 	}
 }
 
+static void ipa_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+}
+
 static void ipa_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	switch (event) {
 	case XUA_ASP_E_M_ASP_DOWN_REQ:
-		ipa_asp_fsm_del_route(fi->priv);
 		osmo_fsm_inst_state_chg(fi, IPA_ASP_S_DOWN, 0, 0);
 		break;
 	}
@@ -1084,17 +1076,6 @@ static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	}
 }
 
-static void ipa_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
-{
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
-}
-
-static void ipa_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
-{
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
-}
-
 static void ipa_pong_timer_cb(void *_fi)
 {
 	struct osmo_fsm_inst *fi = _fi;
@@ -1119,7 +1100,7 @@ static const struct osmo_fsm_state ipa_asp_states[] = {
 	[IPA_ASP_S_DOWN] = {
 		.in_event_mask = S(XUA_ASP_E_M_ASP_UP_REQ) |
 				 S(XUA_ASP_E_SCTP_EST_IND),
-		.out_state_mask = S(IPA_ASP_S_WAIT_ID_ACK) |
+		.out_state_mask = S(IPA_ASP_S_WAIT_ID_GET) |
 				  S(IPA_ASP_S_WAIT_ID_RESP),
 		.name = "ASP_DOWN",
 		.action = ipa_asp_fsm_down,
@@ -1144,8 +1125,10 @@ static const struct osmo_fsm_state ipa_asp_states[] = {
 	},
 	/* Client Side */
 	[IPA_ASP_S_WAIT_ID_GET] = {
-		.in_event_mask = S(IPA_ASP_E_ID_GET),
-		.out_state_mask = S(IPA_ASP_S_WAIT_ID_ACK),
+		.in_event_mask = S(IPA_ASP_E_ID_GET) |
+				 S(IPA_ASP_E_ID_ACK), /* support broken MSCs skipping GET+RESP */
+		.out_state_mask = S(IPA_ASP_S_WAIT_ID_ACK) |
+				  S(IPA_ASP_S_ACTIVE),  /* support broken MSCs skipping GET+RESP */
 		.name = "WAIT_ID_GET",
 		.action = ipa_asp_fsm_wait_id_get,
 	},
@@ -1176,11 +1159,6 @@ static const struct osmo_fsm_state ipa_asp_states[] = {
 	},
 };
 
-static void ipa_asp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
-{
-	ipa_asp_fsm_del_route(fi->priv);
-}
-
 struct osmo_fsm ipa_asp_fsm = {
 	.name = "IPA_ASP",
 	.states = ipa_asp_states,
@@ -1193,7 +1171,6 @@ struct osmo_fsm ipa_asp_fsm = {
 			       S(XUA_ASP_E_ASPSM_BEAT) |
 			       S(XUA_ASP_E_ASPSM_BEAT_ACK),
 	.allstate_action = ipa_asp_allstate,
-	.cleanup = ipa_asp_fsm_cleanup,
 };
 
 
